#!/usr/bin/perl -w

=head1 NAME

vm-backup-audit.pl - A script for identifying which virtual machines have been backed up recently

=head1 SYNOPSIS

vm-backup-audit.pl I<vcenter-server> I<datacenter>

=head1 DESCRIPTION

HP Data Protector includes a report on the last successful backup for
various objects, but it doesn't provide a convenient way of tying that
in with what is on a VMware cluster.

This program queries the I<vcenter-server> given as a command-line
argument, and identifies all the virtual machines on that server.  It
also queries the Data Protector internal database for the last 14 days
to find out what objects have been backed up during VEAgent
backups. It then prepares a list of virtual machine names and shows
when they were last (successfully or unsuccessfully) backed up.

=head1 SAMPLE OUTPUT

If the virtual machine has never been backed up successfully in the
time frame , the message C<no full backup completed cleanly> will be
shown. Otherwise, the relevant session IDs will be shown in reverse
chronological order.

 [LaptopDatacenter:linuxvm1] 2014/12/29-6 (Failed Full) 2014/12/29-5 (Incremental) 2014/12/29-4 (Full)
 [LaptopDatacenter:unbacked-up-vm] <no full backup completed cleanly> 
 [LaptopDatacenter:linuxvm3] 2014/12/29-6 (Failed Full) 2014/12/29-5 (Incremental) 2014/12/29-4 (Full) 
 [LaptopDatacenter:linuxvm1-restored] 2014/12/29-6 (Failed Full) 2014/12/29-5 (Incremental) 2014/12/29-4 (Full) 


=head1 BUGS

This program doesn't correctly handle two virtual machines with the
same name in the same data center. This is probably impossible anyway,
so doesn't matter. If they are in different data centers it is able to
distinguish them.

It's not smart enough to understand that a virtual machine might be
getting cloned or replicated between data centres.

It might not cope very well with mixed Hyper-V and VMware environments.

It might not cope very well with two instances running simultaneously.


=head1 REQUIREMENTS

You must install the VMware Perl API ( L<https://developercenter.vmware.com/web/sdk/55/vsphere-perl> ).

This has only been tested on Linux, and hard-codes the path names to
C<vidiscover.pl> and C<omnidb>.

The cell console must be installed and the cell_server file set up. In
other words C<omnidb> needs to be able to run. (You don't have to run
this on the cell manager, but it might be a little faster if you do.)

There needs to be a .visdkrc file (typically in the current working directory) with lines as follows:

 VI_USERNAME = ....
 VI_PASSWORD = ....

This user should have at least read-only access to the data center. No other permissions are required.



=head1 NOTES

In order to speed up reading, it creates a cache of session files C<.cache->I<session-number>. These can be
deleted at any time.

=cut


# Number of days to search back through
my $timeframe = 14;

my $vcenter = shift;
my $datacenter = shift;

die "Usage: $0 vcenter-server datacenter" unless defined $vcenter and defined $datacenter and $vcenter !~ /^\s*$/ and $datacenter !~ /^\s$/;

use strict;

my $verbose = -t STDOUT;

open(SESSION_HISTORY,"/opt/omni/bin/omnidb -session -type backup -last $timeframe -detail |") || die "Couldn't get history of last $timeframe days";

my %veagent_sessions = ();
my $current_session = undef;
while (<SESSION_HISTORY>) {
  chomp;
#  print STDERR "---> $_\n";
  if (m!SessionID : (\d{4}/\d{2}/\d{2}-\d+)!) {
    $current_session = $1;
    print "Scanning $current_session        \r" if $verbose;
    next;
  }

  next unless /Backup Specification: /;
  if (/Backup Specification: VEAgent (.*)\s*$/) {
    $veagent_sessions{$current_session} = $1;
    #print STDERR "Session $current_session was a VEAgent session ($1)\n";
    $veagent_sessions{$current_session} =~ s/\s*$//;
    $current_session = undef;
    next;
  }
  #print STDERR "Session $current_session was not a VEagent session\n";
  $current_session = undef;
}
close(SESSION_HISTORY);

my %virtual_machines_backed_up = ();

my $current_object = undef;
my $vmpath = undef;
foreach $current_session (keys %veagent_sessions) {
  print STDERR "Analysing $current_session              \r" if $verbose;

  # Since sessions don't change much, I've got a bit of a caching
  # thing happening here.
  
  my $using_cache = 0;
  my $flatname = ".cache-$current_session";
  $flatname =~ s:/:-:g;
  if (-e $flatname) { $using_cache = 1; }
  if ($using_cache) { $using_cache = 0 unless open(SESSION_DETAILS,$flatname); }
  if (!$using_cache) {
    my $devnull = $verbose ? " 2> /dev/null " : "";
    open(SESSION_DETAILS,"/opt/omni/bin/omnidb -session $current_session -detail $devnull |");
    unless (open(CACHE,">$flatname")) {
      print STDERR "\n$! writing to cache $flatname\n";
      $flatname = undef;
    }
  }
  print STDERR "Reading $current_session from cache  \r" if $verbose and $using_cache;
  print STDERR "Reading $current_session, no caching  \r" if $verbose and !defined $flatname;
  print STDERR "Reading $current_session, caching     \r" if $verbose and defined $flatname;
  while (<SESSION_DETAILS>) {
    if (!$using_cache && defined $flatname) { print CACHE $_; }
    chomp;
    #print STDERR " --> $_\n";
    if (/^Object name : ([^:]*):(.*)/) {
      if ((defined $current_object) && (exists $current_object->{"VM Path"})) {
	$vmpath = $current_object->{"VM Path"};
	unless (exists $virtual_machines_backed_up{$vmpath}) {
	  $virtual_machines_backed_up{$vmpath} = [];
	}
	push(@{$virtual_machines_backed_up{$vmpath}},$current_object);
      }
      $current_object = {};
      $current_object->{"vcenter"} = $1;
      $current_object->{"dppathname"} = $2;
      #print STDERR " Got an object for $1:$2\n";
      next;
    }
    next unless defined $current_object;
#    print STDERR " Investigating $_\n";
    next unless /^\s*([^:]*)\s*: (.*)\s*$/;
    my $k = $1;
    my $v = $2;
    $k =~ s/^\s*//;
    $k =~ s/\s*$//;
    $v =~ s/^\s*//;
    $v =~ s/\s*$//;
#    print STDSERR "Bingo.\n";
    $current_object->{$k} = $v;
  }
  close(SESSION_DETAILS);
  if ((defined $current_object) && (exists $current_object->{"vmpath"})) {
    $vmpath = $current_object->{"VM Path"};
    unless (exists $virtual_machines_backed_up{$vmpath}) {
      $virtual_machines_backed_up{$vmpath} = [];
    }
    push(@{$virtual_machines_backed_up{$vmpath}},$current_object);
  }
  if (!$using_cache && defined $flatname) { close(CACHE); }
}

sub bysession {
  my ($a_year,$a_month,$a_extra) = split(m:/:,$a->{"Backup ID"});
  my ($b_year,$b_month,$b_extra) = split(m:/:,$b->{"Backup ID"});
  my ($a_day,$a_num) = split(/-/,$a_extra);
  my ($b_day,$b_num) = split(/-/,$b_extra);
  if ($a_year != $b_year) { return $a_year <=> $b_year; }
  if ($a_month != $b_month) { return $a_month <=> $b_month; }
  if ($a_day != $b_day) { return $a_day <=> $b_day; }
  return $a_num <=> $b_num;
}





my %already_reported_virtual_machines = ();
my $vmname;
open(VIDISCOVERY,"perl /usr/lib/vmware-vcli/apps/general/vidiscovery.pl --url https://$vcenter:443/sdk/vimService --managedentity datacenter --entityname $datacenter | ") || die "can't run vidiscovery.pl";
while (<VIDISCOVERY>) {
  if (/^\s*VM : (.*)$/) {
    my $vmname = $1;
    $vmname =~ s/\s*$//;
    $vmname =~ s/^\s*$//;
    next if exists $already_reported_virtual_machines{$vmname};
    $already_reported_virtual_machines{$vmname} = 1;
    my $done_full = 0;
    my $done_incr = 0;
    my $done_diff = 0;
    print "[$datacenter:$vmname] ";
    foreach $vmpath (sort keys %virtual_machines_backed_up) {
      next unless $vmpath =~ m:$datacenter/.*/$vmname:;
      #print STDERR "$vmpath: " if $verbose;
      foreach $current_object (reverse sort bysession @{$virtual_machines_backed_up{$vmpath}}) {
	my $status = $current_object->{"Object status"};
	my $type = $current_object->{"Backup type"};
	my $current_session = $current_object->{"Backup ID"};
	if ($status eq 'Pending' or $status eq 'Running') {
	  # Not only should we ignore this, but we should remove the cache file
	  my $flatname = ".cache-$current_session";
	  $flatname =~ s:/:-:g;
	  unlink $flatname if -e $flatname;
	  next;
	}
	if ($status eq 'Completed') {
	  if ($type eq 'Full') {
	    next if $done_full;
	    $done_full = 1;
	    print "$current_session ($type) ";
	    next;
	  }
	  if ($type eq 'Incremental') {
	    next if $done_incr;
	    $done_incr = 1;
	    print "$current_session ($type) ";
	    next;
	  }
	  if ($type eq 'Differential') {
	    next if $done_diff;
	    $done_diff = 1;
	    print "$current_session ($type) ";
	    next;
	  }
	}
	print "$current_session ($status $type) ";	
      }
    }
    if (!$done_full) { print "<no full backup completed cleanly> "; }
    print "\n";
  }
}

close(VIDISCOVERY);


